@@ -19,6 +19,10 @@
 # FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 # IN THE SOFTWARE.
 
+"""
+Note: The `docspec_python.parser` module is not public API.
+"""
+
 import dataclasses
 import os
 import re
@@ -26,13 +30,17 @@ import textwrap
 import typing as t
 
 from docspec import (
+  ApiObject,
   Argument,
   Class,
   Data,
+  Docstring,
   Decoration,
   Function,
+  Indirection,
   Location,
-  Module)
+  Module,
+  _ModuleMembers)
 from lib2to3.refactor import RefactoringTool  # type: ignore
 from lib2to3.pgen2 import token
 from lib2to3.pgen2.parse import ParseError
@@ -102,20 +110,22 @@ class Parser:
 
     for node in ast.children:
       member = self.parse_declaration(module, node)
-      if member:
+      if isinstance(member, list):
+        module.members += member
+      elif member:
         module.members.append(member)
 
+    module.sync_hierarchy()
     return module
 
-  def parse_declaration(self, parent, node, decorations=None):
+  def parse_declaration(self, parent, node, decorations=None) -> t.Union[None, _ModuleMembers, t.List[_ModuleMembers]]:
     if node.type == syms.simple_stmt:
       assert not decorations
       stmt = node.children[0]
-      if stmt.type in (syms.import_name, syms.import_from):
-        # TODO @NiklasRosenstein handle import statements?
-        pass
+      if stmt.type in (syms.import_stmt, syms.import_name, syms.import_from, syms.import_as_names, syms.import_as_name):
+        return list(self.parse_import(node, stmt))
       elif stmt.type == syms.expr_stmt:
-        return self.parse_statement(parent, stmt)
+        return self.parse_statement(node, stmt)
     elif node.type == syms.funcdef:
       return self.parse_funcdef(parent, node, False, decorations)
     elif node.type == syms.classdef:
@@ -165,6 +175,55 @@ class Parser:
     result = dict(_parse([], ('names', []), stmt))
     return result.get('names', []), result.get('annotation', []), result.get('value', [])
 
+  def parse_import(self, parent, node: Node) -> t.Iterable[Indirection]:
+
+    def _single_import_to_indirection(node: t.Union[Node, Leaf]) -> Indirection:
+      if node.type == syms.dotted_as_name:  # example: urllib.request as r
+        target = self.name_to_string(node.children[0])
+        name = self.name_to_string(node.children[2])
+        return Indirection(name, self.location_from(node), None, target)
+      elif node.type == syms.dotted_name:  # example os.path
+        name = self.name_to_string(node)
+        return Indirection(name.split('.')[-1], self.location_from(node), None, name)
+      elif isinstance(node, Leaf):
+        return Indirection(node.value, self.location_from(node), None, node.value)
+      else:
+        raise RuntimeError(f'cannot handle {node!r}')
+
+    def _from_import_to_indirection(prefix: str, node: t.Union[Node, Leaf]) -> Indirection:
+      if node.type == syms.import_as_name:  # example: Widget as W
+        target = self.name_to_string(node.children[0])
+        name = self.name_to_string(node.children[2])
+        return Indirection(name, self.location_from(node), None, prefix + '.' + target)
+      elif isinstance(node, Leaf):  # example: Widget
+        name = self.name_to_string(node)
+        if not prefix.endswith('.'):
+          prefix += '.'
+        return Indirection(name, self.location_from(node), None, prefix + name)
+      else:
+        raise RuntimeError(f'cannot handle {node!r}')
+
+    if node.type == syms.import_name:  # example: import ...
+      subject_node = node.children[1]
+      if subject_node.type == syms.dotted_as_names:
+        yield from (_single_import_to_indirection(n) for n in subject_node.children if n.type != token.COMMA)
+      else:
+        yield _single_import_to_indirection(subject_node)
+
+    elif node.type == syms.import_from:  # example: from xyz import ...
+      index = next(i for i, n in enumerate(node.children) if isinstance(n, Leaf) and n.type == token.NAME and n.value == 'import')
+      name = ''.join(self.name_to_string(x) for x in node.children[1:index])
+      subject_node = node.children[index + 1]
+      if subject_node.type == token.LPAR:
+        subject_node = node.children[index + 2]
+      if subject_node.type == syms.import_as_names:
+        yield from (_from_import_to_indirection(name, n) for n in subject_node.children if n.type not in (token.LPAR, token.RPAR, token.COMMA))
+      else:
+        yield _from_import_to_indirection(name, subject_node)
+
+    else:
+      raise RuntimeError(f'dont know how to deal with {node!r}')
+
   def parse_statement(self, parent, stmt):
     names, annotation, value = self._split_statement(stmt)
     if value or annotation:
@@ -345,6 +404,7 @@ class Parser:
     for child in suite.children if suite else []:
       if isinstance(child, Node):
         member = self.parse_declaration(class_, child)
+        assert not isinstance(member, (list, Module)), member
         if metaclass is None and isinstance(member, Data) and \
             member.name == '__metaclass__':
           metaclass = member.value
@@ -354,17 +414,17 @@ class Parser:
     class_.metaclass = metaclass
     return class_
 
-  def location_from(self, node):
+  def location_from(self, node: t.Union[Node, Leaf]) -> Location:
     return Location(self.filename, node.get_lineno())
 
-  def get_return_annotation(self, node):
+  def get_return_annotation(self, node: Node) -> t.Optional[str]:
     rarrow = find(lambda x: x.type == token.RARROW, node.children)
     if rarrow:
       node = rarrow.next_sibling
       return self.nodes_to_string([node])
     return None
 
-  def get_most_recent_prefix(self, node):
+  def get_most_recent_prefix(self, node) -> str:
     if node.prefix:
       return node.prefix
     while not node.prev_sibling and not node.prefix:
@@ -378,7 +438,7 @@ class Parser:
       node = node.children[-1]
     return node.prefix
 
-  def get_docstring_from_first_node(self, parent, module_level=False):
+  def get_docstring_from_first_node(self, parent: Node, module_level: bool = False) -> t.Optional[Docstring]:
     """
     This method retrieves the docstring for the block node *parent*. The
     node either declares a class or function.
@@ -387,7 +447,7 @@ class Parser:
     assert parent is not None
     node = find(lambda x: isinstance(x, Node), parent.children)
     if node and node.type == syms.simple_stmt and node.children[0].type == token.STRING:
-      return self.prepare_docstring(node.children[0].value)
+      return self.prepare_docstring(node.children[0].value, parent)
     if not node and not module_level:
       return None
     if self.options.treat_singleline_comment_blocks_as_docstrings:
@@ -396,7 +456,7 @@ class Parser:
         return docstring
     return None
 
-  def get_statement_docstring(self, node):
+  def get_statement_docstring(self, node: Node) -> t.Optional[Docstring]:
     prefix = self.get_most_recent_prefix(node)
     match = re.match(r'\s*', prefix[::-1])
     assert match is not None
@@ -406,15 +466,17 @@ class Parser:
       if doc_type == 'statement':
         return docstring
     # Look for the next string literal instead.
-    while node and node.type != syms.simple_stmt:
-      node = node.parent
-    if node and node.next_sibling and node.next_sibling.type == syms.simple_stmt:
-      string_literal = node.next_sibling.children[0]
+    curr: t.Optional[Node] = node
+    while curr and curr.type != syms.simple_stmt:
+      curr = curr.parent
+    if curr and curr.next_sibling and curr.next_sibling.type == syms.simple_stmt:
+      string_literal = curr.next_sibling.children[0]
       if string_literal.type == token.STRING:
-        return self.prepare_docstring(string_literal.value)
+        assert isinstance(string_literal, Leaf)
+        return self.prepare_docstring(string_literal.value, string_literal)
     return None
 
-  def get_hashtag_docstring_from_prefix(self, node: Node) -> t.Tuple[t.Optional[str], t.Optional[str]]:
+  def get_hashtag_docstring_from_prefix(self, node: Node) -> t.Tuple[t.Optional[Docstring], t.Optional[str]]:
     """
     Given a node in the AST, this method retrieves the docstring from the
     closest prefix of this node (ie. any block of single-line comments that
@@ -440,12 +502,15 @@ class Parser:
         doc_type = 'block'
       if lines or line:
         lines.append(line)
-    return self.prepare_docstring('\n'.join(reversed(lines))), doc_type
 
-  def prepare_docstring(self, s):
+    return self.prepare_docstring('\n'.join(reversed(lines)), node), doc_type
+
+  def prepare_docstring(self, s: str, node_for_location: t.Union[Node, Leaf]) -> t.Optional[Docstring]:
     # TODO @NiklasRosenstein handle u/f prefixes of string literal?
+    location = self.location_from(node_for_location)
     s = s.strip()
     if s.startswith('#'):
+      location.lineno -= s.count('\n') + 2
       lines = []
       for line in s.split('\n'):
         line = line.strip()
@@ -454,11 +519,11 @@ class Parser:
         else:
           line = line[1:]
         lines.append(line.lstrip())
-      return '\n'.join(lines).strip()
+      return Docstring('\n'.join(lines).strip(), location)
     if s.startswith('"""') or s.startswith("'''"):
-      return dedent_docstring(s[3:-3]).strip()
+      return Docstring(dedent_docstring(s[3:-3]).strip(), location)
     if s.startswith('"') or s.startswith("'"):
-      return dedent_docstring(s[1:-1]).strip()
+      return Docstring(dedent_docstring(s[1:-1]).strip(), location)
     return None
 
   def nodes_to_string(self, nodes):
